// Package configfile provides the utilities for our config file parsing.
//
// Note the API here is not considered stable and can and will change and we see fit.
// The purpose is to use this only for our own config file parsing such as
// containers.conf, storage.conf and registries.conf. We will not consider use cases
// for external consumers.
package configfile
